"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSONObject = exports.JSONTypeError = void 0;
exports.validate = validate;
exports.custom = custom;
exports.union = union;
exports.required = required;
exports.optional = optional;
exports.map = map;
exports.passthrough = passthrough;
exports.integer = integer;
exports.gt = gt;
exports.gte = gte;
exports.lt = lt;
exports.lte = lte;
exports.eq = eq;
exports.ne = ne;
exports.array = array;
exports.maxLength = maxLength;
exports.minLength = minLength;
require("reflect-metadata");
let __optional = Symbol('optional');
let __keys = Symbol('keys');
let __mapping = Symbol('mapping');
let __union = Symbol('union');
let __code = Symbol('code');
let __gt = Symbol('gt');
let __lt = Symbol('lt');
let __eq = Symbol('eq');
let __ne = Symbol('ne');
let __gte = Symbol('gte');
let __lte = Symbol('lte');
let __default = Symbol('default');
let __passthrough = Symbol('passthrough');
let __integer = Symbol('integer');
let __array = Symbol('array');
let __custom = Symbol('custom');
let __minLength = Symbol('minLength');
let __maxLength = Symbol('maxLength');
class JSONTypeError extends TypeError {
}
exports.JSONTypeError = JSONTypeError;
class JSONObject extends Object {
    newValue(value, design_type, key) {
        if (typeof value == 'object') {
            if (Array.isArray(value)) {
                let array_meta = Reflect.hasMetadata(__array, this, key) ? Reflect.getMetadata(__array, this, key) : null;
                if (array_meta && value.length) {
                    let ret = Array();
                    for (let i = 0; i < value.length; ++i) {
                        let e = value[i];
                        let new_e;
                        if (typeof e == 'object') {
                            if (Array.isArray(e)) { // (at least) 2 dim array, throw an unsupported error
                                throw new JSONTypeError(`${this.constructor.name}.${key}: array type cannot be array, consider using @passthrough and @validate`);
                            }
                            new_e = new array_meta(e);
                        }
                        else {
                            new_e = array_meta(e);
                        }
                        let expected_type = typeof e;
                        let value_type = typeof new_e;
                        if (expected_type != value_type) {
                            throw new JSONTypeError(`${this.constructor.name}.${key} array element requires type '${expected_type}', got '${value_type}' instead`);
                        }
                        ret.push(new_e);
                    }
                    return ret;
                }
                else {
                    return value;
                }
            }
            else if (design_type.name == 'Array') {
                if (value != null) {
                    throw new JSONTypeError(`${this.constructor.name}.${key}: expected array type, got '${typeof value}' instead`);
                }
            }
            else {
                return new design_type(value);
            }
            //
        }
        else if (design_type.name == 'Date') {
            return new design_type(value);
        }
        else {
            return design_type(value);
        }
    }
    set(key, value) {
        let new_value;
        let design_type = Reflect.getMetadata("design:type", this, key);
        if (design_type === undefined) {
            throw new Error("Meta data not defined");
        }
        if (Reflect.hasMetadata(__union, this, key)) {
            let values = (Reflect.getMetadata(__union, this, key));
            let is_valid = false;
            for (let i = 0; i < values.length && !is_valid; ++i) {
                if (value === values[i]) {
                    is_valid = true;
                }
            }
            if (!is_valid) {
                throw new JSONTypeError(`${this.constructor.name}.${key} requires one of the following values:[${values}], got '${value}' instead`);
            }
        }
        if (Reflect.hasMetadata(__passthrough, this, key) && Reflect.getMetadata(__passthrough, this, key) == true) {
            new_value = value;
        }
        else {
            // Run custom assignment if exists
            if (Reflect.hasMetadata(__custom, this, key)) {
                let code = (Reflect.getMetadata(__custom, this, key));
                let ret = code(this, key, value);
                if (ret !== undefined)
                    new_value = ret;
            }
            else {
                new_value = this.newValue(value, design_type, key);
                // If not date
                if (design_type.name == 'Date') {
                    if (typeof new_value !== 'object' || typeof new_value.getTime !== 'function') {
                        throw new JSONTypeError(`${this.constructor.name}.${key} requires a type that can be converted to a date, got '${typeof value}' instead`);
                    }
                    if (isNaN(new_value.getTime())) {
                        throw new JSONTypeError(`${this.constructor.name}.${key} requires a type that can be converted to a date, got '${typeof value}(${value})' instead`);
                    }
                }
                else {
                    let expected_type = typeof new_value;
                    let value_type = typeof value;
                    if (expected_type != value_type) {
                        throw new JSONTypeError(`${this.constructor.name}.${key} requires type '${expected_type}', got '${value_type}' instead`);
                    }
                }
                // Run validations if exists
                let symbols = [__code, __gt, __gte, __eq, __ne, __lt, __lte, __minLength, __maxLength];
                for (let symbol of symbols) {
                    if (Reflect.hasMetadata(symbol, this, key)) {
                        let code = (Reflect.getMetadata(symbol, this, key));
                        code(this, key, new_value);
                    }
                }
                // Run integer validator
                if (Reflect.hasMetadata(__integer, this, key)) {
                    this.validateNumeric(key, new_value);
                    if (new_value != Math.floor(new_value)) {
                        throw new JSONTypeError(`${this.constructor.name}.${key} should be an integer value, got ${new_value} instead`);
                    }
                }
            }
        }
        this[key] = new_value;
    }
    get(key) {
        return this[key];
    }
    constructor(json) {
        super();
        if (json === undefined) {
            json = {};
        }
        let json_keys = JSONObject.jsonKeys(this);
        if (!json_keys)
            return; // no properties defined
        for (let key of json_keys) {
            let json_value;
            let json_key;
            // Is there a mapping this value?
            if (Reflect.hasMetadata(__mapping, this, key)) {
                json_key = Reflect.getMetadata(__mapping, this, key);
            }
            else {
                json_key = key;
            }
            json_value = json[json_key];
            // Is there a default?
            if (json_value === undefined || json_value === null) {
                let default_json_value = Reflect.getMetadata(__default, this, key);
                // Only set default value if it's there, otherwise leave the value null or undefiend (there's a difference between the two)
                if (default_json_value !== undefined) {
                    json_value = default_json_value;
                }
            }
            // Still undefined?
            if (json_value === undefined || json_value === null) {
                if (this.get(key) === undefined && !JSONObject.isOptional(this, key)) {
                    throw new JSONTypeError(`${this.constructor.name}.${key} is required`);
                }
                this[key] = json_value; // undefined or null, exactly as specified (or not specified :-) )
            }
            else {
                this.set(key, json_value);
            }
        }
    }
    static metadata(symbol, value) {
        let meta = Reflect.metadata(symbol, value);
        let ret = (target, key) => {
            JSONObject.preprocess(target, key);
            meta(target, key);
        };
        return ret;
    }
    static custom(code) {
        return JSONObject.metadata(__custom, code);
    }
    static validate(code) {
        return JSONObject.metadata(__code, code);
    }
    static union(values) {
        return JSONObject.metadata(__union, values);
    }
    static map(newKey) {
        return JSONObject.metadata(__mapping, newKey);
    }
    static isOptional(target, key) {
        let ret = Reflect.getMetadata(__optional, target, key);
        return ret == true || ret === undefined;
    }
    static jsonKeys(target) {
        let ret = Array();
        // Get the keys, all the way to the base class
        while (target = target.__proto__) {
            let i = Reflect.getMetadata(__keys, target, target.constructor.name);
            if (i) {
                ret = ret.concat(i);
            }
        }
        return ret;
    }
    static preprocess(target, key) {
        let keys;
        if (Reflect.hasMetadata(__keys, target, target.constructor.name)) {
            keys = Reflect.getMetadata(__keys, target, target.constructor.name);
        }
        else {
            keys = Array();
            Reflect.defineMetadata(__keys, keys, target, target.constructor.name);
        }
        if (!keys.length || keys[keys.length - 1] != key) {
            keys.push(key);
        }
    }
    validateNumeric(key, value) {
        if (typeof value !== 'number') {
            throw new JSONTypeError(`${this.constructor.name}.${key}: requires numeric type for validation operators`);
        }
    }
    static array(type) {
        return JSONObject.metadata(__array, type);
    }
    static integer(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__integer, true, target, key);
    }
    static gt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue > capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} > ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__gt, validator);
    }
    static gte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue >= capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} >= ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__gte, validator);
    }
    static lt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue < capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} < ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__lt, validator);
    }
    static lte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue <= capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} <= ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__lte, validator);
    }
    static eq(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue == capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} == ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__eq, validator);
    }
    static ne(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue != capture)) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue} != ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__ne, validator);
    }
    static passthrough(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__passthrough, true, target, key);
    }
    static required(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, false, target, key);
    }
    static optional(target, key) {
        if (key === undefined) {
            let meta = Reflect.metadata(__default, target);
            let ret = (target, key) => {
                JSONObject.optional(target, key); // Make it optional
                meta(target, key);
            };
            return ret;
        }
        else {
            JSONObject.preprocess(target, key);
            Reflect.defineMetadata(__optional, true, target, key);
        }
    }
    static minLength(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (jsonValue.length < value) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue}.length >= ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__minLength, validator);
    }
    static maxLength(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (jsonValue.length > value) {
                throw new JSONTypeError(`${object.constructor.name}.${key}: ${jsonValue}.length <= ${capture} requirement failed`);
            }
        };
        return JSONObject.metadata(__maxLength, validator);
    }
}
exports.JSONObject = JSONObject;
function validate(code) {
    return JSONObject.validate(code);
}
function custom(code) {
    return JSONObject.custom(code);
}
function union(values) {
    return JSONObject.union(values);
}
function required(target, key) {
    JSONObject.required(target, key);
}
function optional(target, key) {
    if (key === undefined) {
        return JSONObject.optional(target, key);
    }
    else {
        JSONObject.optional(target, key);
    }
}
function map(jsonKey) {
    return JSONObject.map(jsonKey);
}
function passthrough(target, key) {
    return JSONObject.passthrough(target, key);
}
function integer(target, key) {
    return JSONObject.integer(target, key);
}
function gt(value) {
    return JSONObject.gt(value);
}
function gte(value) {
    return JSONObject.gte(value);
}
function lt(value) {
    return JSONObject.lt(value);
}
function lte(value) {
    return JSONObject.lte(value);
}
function eq(value) {
    return JSONObject.eq(value);
}
function ne(value) {
    return JSONObject.ne(value);
}
function array(value) {
    return JSONObject.array(value);
}
function maxLength(value) {
    return JSONObject.maxLength(value);
}
function minLength(value) {
    return JSONObject.minLength(value);
}
//# sourceMappingURL=json_object.js.map